"""Trust Region Reflective algorithm for least-squares optimization.
The algorithm is based on ideas from paper [STIR]_. The main idea is to
account for the presence of the bounds by appropriate scaling of the variables (or,
equivalently, changing a trust-region shape). Let's introduce a vector v::

           | ub[i] - x[i], if g[i] < 0 and ub[i] < np.inf
    v[i] = | x[i] - lb[i], if g[i] > 0 and lb[i] > -np.inf
           | 1,           otherwise

where g is the gradient of a cost function and lb, ub are the bounds. Its
components are distances to the bounds at which the anti-gradient points (if
this distance is finite). Define a scaling matrix D = diag(v**0.5).
First-order optimality conditions can be stated as::

    D^2 g(x) = 0.

Meaning that components of the gradient should be zero for strictly interior
variables, and components must point inside the feasible region for variables
on the bound.
Now consider this system of equations as a new optimization problem. If the
point x is strictly interior (not on the bound), then the left-hand side is
differentiable and the Newton step for it satisfies::

    (D^2 H + diag(g) Jv) p = -D^2 g

where H is the Hessian matrix (or its J^T J approximation in least squares),
Jv is the Jacobian matrix of v with components -1, 1 or 0, such that all
elements of matrix C = diag(g) Jv are non-negative. Introduce the change
of the variables x = D x_h (_h would be "hat" in LaTeX). In the new variables,
we have a Newton step satisfying::

    B_h p_h = -g_h,

where B_h = D H D + C, g_h = D g. In least squares B_h = J_h^T J_h, where
J_h = J D. Note that J_h and g_h are proper Jacobian and gradient with respect
to "hat" variables. To guarantee global convergence we formulate a
trust-region problem based on the Newton step in the new variables::

    0.5 * p_h^T B_h p + g_h^T p_h -> min, ||p_h|| <= Delta

In the original space B = H + D^{-1} C D^{-1}, and the equivalent trust-region
problem is::

    0.5 * p^T B p + g^T p -> min, ||D^{-1} p|| <= Delta

Here, the meaning of the matrix D becomes more clear: it alters the shape
of a trust-region, such that large steps towards the bounds are not allowed.
In the implementation, the trust-region problem is solved in "hat" space,
but handling of the bounds is done in the original space (see below and read
the code).
The introduction of the matrix D doesn't allow to ignore bounds, the algorithm
must keep iterates strictly feasible (to satisfy aforementioned
differentiability), the parameter theta controls step back from the boundary
(see the code for details).
The algorithm does another important trick. If the trust-region solution
doesn't fit into the bounds, then a reflected (from a firstly encountered
bound) search direction is considered. For motivation and analysis refer to
[STIR]_ paper (and other papers of the authors). In practice, it doesn't need
a lot of justifications, the algorithm simply chooses the best step among
three: a constrained trust-region step, a reflected step and a constrained
Cauchy step (a minimizer along -g_h in "hat" space, or -D^2 g in the original
space).
Another feature is that a trust-region radius control strategy is modified to
account for appearance of the diagonal C matrix (called diag_h in the code).
Note that all described peculiarities are completely gone as we consider
problems without bounds (the algorithm becomes a standard trust-region type
algorithm very similar to ones implemented in MINPACK).
The implementation supports two methods of solving the trust-region problem.
The first, called 'exact', applies SVD on Jacobian and then solves the problem
very accurately using the algorithm described in [JJMore]_. It is not
applicable to large problem. The second, called 'lsmr', uses the 2-D subspace
approach (sometimes called "indefinite dogleg"), where the problem is solved
in a subspace spanned by the gradient and the approximate Gauss-Newton step
found by ``scipy.sparse.linalg.lsmr``. A 2-D trust-region problem is
reformulated as a 4th order algebraic equation and solved very accurately by
``numpy.roots``. The subspace approach allows to solve very large problems
(up to couple of millions of residuals on a regular PC), provided the Jacobian
matrix is sufficiently sparse.
References
----------
.. [STIR] Branch, M.A., T.F. Coleman, and Y. Li, "A Subspace, Interior,
      and Conjugate Gradient Method for Large-Scale Bound-Constrained
      Minimization Problems," SIAM Journal on Scientific Computing,
      Vol. 21, Number 1, pp 1-23, 1999.
.. [JJMore] More, J. J., "The Levenberg-Marquardt Algorithm: Implementation
    and Theory," Numerical Analysis, ed. G. A. Watson, Lecture
"""
import numpy as np
from numpy.linalg import norm
import time
from typing import Callable, Optional, Tuple, Union, List, Dict, Any, Sequence

from jax.config import config
config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.scipy.linalg import svd as jax_svd
from jax import jit
from jax.tree_util import tree_flatten

from jaxfit.common_scipy import (update_tr_radius, solve_lsq_trust_region, 
                    check_termination, CL_scaling_vector,
                    make_strictly_feasible, find_active_constraints, in_bounds, 
                    step_size_to_bound, intersect_trust_region, 
                    minimize_quadratic_1d, print_header_nonlinear,
                    print_iteration_nonlinear)

from jaxfit.common_jax import CommonJIT
from jaxfit._optimize import OptimizeResult


class TrustRegionJITFunctions():
    """JIT functions for trust region algorithm."""
    
    def __init__(self):
        """Call all of the create functions which create the JAX/JIT functions
        that are members of the class."""
        self.create_grad_func()
        self.create_grad_hat()
        self.create_svd_funcs()
        self.create_default_loss_func()
        self.create_calculate_cost()
        self.create_check_isfinite()


    def create_default_loss_func(self):
        """Create the default loss function which is simply the sum of the
        squares of the residuals."""
        @jit
        def loss_function(f: jnp.ndarray) -> jnp.ndarray:
            """The default loss function is the sum of the squares of the
            residuals divided by two.

            Parameters
            ----------
            f : jnp.ndarray
                The residuals.

            Returns
            -------
            jnp.ndarray
                The loss function value.
            """
            return 0.5 * jnp.dot(f, f)
            # return 0.5 * jnp.sum(f**2)
            
        self.default_loss_func = loss_function

    
    def create_grad_func(self):
        """Create the function to compute the gradient of the loss function
        which is simply the function evaluation dotted with the Jacobian."""
        @jit
        def compute_grad(J: jnp.ndarray, f: jnp.ndarray) -> jnp.ndarray:
            """Compute the gradient of the loss function.

            Parameters
            ----------
            J : jnp.ndarray
                The Jacobian matrix.
            f : jnp.ndarray
                The residuals.

            Returns
            -------
            jnp.ndarray
                The gradient of the loss function.
            """
            return f.dot(J)
        self.compute_grad = compute_grad


    def create_grad_hat(self):
        """Calculate the gradient in the "hat" space, which is just multiplying
        the gradient by the diagonal matrix D. This is used in the trust region
        algorithm. Here we only use the diagonals of D, since D is diagonal.
        """
        @jit
        def compute_grad_hat(g: jnp.ndarray, d: jnp.ndarray) -> jnp.ndarray:
            """Compute the gradient in the "hat" space.
            
            Parameters
            ----------
            g : jnp.ndarray
            The gradient of the loss function.
            d : jnp.ndarray
            The diagonal of the diagonal matrix D.

            Returns
            -------
            jnp.ndarray
            The gradient in the "hat" space.
            """
            return d * g
        self.compute_grad_hat = compute_grad_hat

    
    def create_svd_funcs(self):
        """Create the functions to compute the SVD of the Jacobian matrix.
        There are two versions, one for problems with bounds and one for
        problems without bounds. The version for problems with bounds is
        slightly more complicated."""

        @jit
        def svd_no_bounds(J: jnp.ndarray, 
                          d: jnp.ndarray, 
                          f: jnp.ndarray
                          ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, 
                                     jnp.ndarray, jnp.ndarray]:
            """Compute the SVD of the Jacobian matrix, J, in the "hat" space.
            This is the version for problems without bounds.

            Parameters
            ----------
            J : jnp.ndarray
                The Jacobian matrix.
            d : jnp.ndarray
                The diagonal of the diagonal matrix D.
            f : jnp.ndarray
                The residuals.

            Returns
            -------
            J_h : jnp.ndarray
                  the Jacobian matrix in the "hat" space.
            U : jnp.ndarray
                the left singular vectors of the SVD of J_h.
            s : jnp.ndarray
                the singular values of the SVD of J_h.
            V : jnp.ndarray
                the right singular vectors of the SVD of J_h.
            uf : jnp.ndarray
                 the dot product of U.T and f.
            """
            J_h = J * d
            U, s, V = jax_svd(J_h, full_matrices=False)
            V = V.T
            uf = U.T.dot(f)
            return J_h, U, s, V, uf
    
        @jit
        def svd_bounds(f: jnp.ndarray,
                       J: jnp.ndarray,
                       d: jnp.ndarray,
                       J_diag: jnp.ndarray,
                       f_zeros: jnp.ndarray
                       ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, 
                                  jnp.ndarray, jnp.ndarray]:
            """Compute the SVD of the Jacobian matrix, J, in the "hat" space.
            This is the version for problems with bounds.

            Parameters
            ----------
            J : jnp.ndarray
                The Jacobian matrix.
            d : jnp.ndarray
                The diagonal of the diagonal matrix D.
            f : jnp.ndarray
                The residuals.
            J_diag : jnp.ndarray
                    Added term to Jacobian matrix.
            f_zeros : jnp.ndarray
                    Empty residuals for the added term.


            Returns
            -------
            J_h : jnp.ndarray
                  the Jacobian matrix in the "hat" space.
            U : jnp.ndarray
                the left singular vectors of the SVD of J_h.
            s : jnp.ndarray
                the singular values of the SVD of J_h.
            V : jnp.ndarray
                the right singular vectors of the SVD of J_h.
            uf : jnp.ndarray
                 the dot product of U.T and f.
            """
            J_h = J * d
            J_augmented = jnp.concatenate([J_h, J_diag])
            f_augmented = jnp.concatenate([f, f_zeros])
            U, s, V = jax_svd(J_augmented, full_matrices=False)
            V = V.T
            uf = U.T.dot(f_augmented)
            return J_h, U, s, V, uf
        
        self.svd_no_bounds = svd_no_bounds
        self.svd_bounds = svd_bounds
                
            
    def create_calculate_cost(self):
        """Create the function to calculate the cost function."""
        @jit
        def calculate_cost(rho, data_mask):
            """Calculate the cost function.

            Parameters
            ----------
            rho : jnp.ndarray
                The per element cost times two.
            data_mask : jnp.ndarray
                The mask for the data.

            Returns
            -------
            jnp.ndarray
                The cost function.
            """
            cost_array = jnp.where(data_mask, rho[0], 0)
            return 0.5 * jnp.sum(cost_array)
        self.calculate_cost = calculate_cost
    
    
    def create_check_isfinite(self):
        """Create the function to check if the evaluated residuals are finite."""
        @jit
        def isfinite(f_new: jnp.ndarray) -> bool:
            """Check if the evaluated residuals are finite.

            Parameters
            ----------
            f_new : jnp.ndarray
                The evaluated residuals.

            Returns
            -------
            bool
                True if all residuals are finite, False otherwise.
            """
            return jnp.all(jnp.isfinite(f_new))
        self.check_isfinite = isfinite


class TrustRegionReflective(TrustRegionJITFunctions):
    
    def __init__(self):
        """Initialize the TrustRegionReflective class."""
        super().__init__()
        self.cJIT = CommonJIT()


    def trf(self, 
            fun: Callable, 
            xdata: Union[jnp.ndarray, Tuple[jnp.ndarray]], 
            ydata: jnp.ndarray, 
            jac: Callable,
            data_mask: jnp.ndarray, 
            transform: jnp.ndarray, 
            x0: np.ndarray, 
            f0: jnp.ndarray, 
            J0: jnp.ndarray,
            lb: np.ndarray, 
            ub: np.ndarray, 
            ftol: float, 
            xtol: float, 
            gtol: float, 
            max_nfev: int, 
            f_scale: float, 
            x_scale: np.ndarray, 
            loss_function: Union[None, Callable],
            tr_options: Dict, 
            verbose: int, 
            timeit: bool=False
            ) -> Dict:
        """Minimize a scalar function of one or more variables using the
        trust-region reflective algorithm. Although I think this is not good
        coding style, I maintained the original code format from SciPy such
        that the code is easier to compare with the original. See the note
        from the algorithms original author below.
        
        
        For efficiency, it makes sense to run 
        the simplified version of the algorithm when no bounds are imposed. 
        We decided to write the two separate functions. It violates the DRY 
        principle, but the individual functions are kept the most readable.
        
        Parameters
        ----------
        fun : callable
            The residual function
        xdata : array_like or tuple of array_like
            The independent variable where the data is measured. If `xdata` is a
            tuple, then the input arguments to `fun` are assumed to be
            ``(xdata[0], xdata[1], ...)``.
        ydata : jnp.ndarray
            The dependent data
        jac : callable
            The Jacobian of `fun`.
        data_mask : jnp.ndarray
            The mask for the data.
        transform : jnp.ndarray
            The uncertainty transform for the data.
        x0 : jnp.ndarray
            Initial guess. Array of real elements of size (n,), where 'n' is the
            number of independent variables.
        f0 : jnp.ndarray
            Initial residuals. Array of real elements of size (m,), where 'm' is
            the number of data points.
        J0 : jnp.ndarray
            Initial Jacobian. Array of real elements of size (m, n), where 'm' is
            the number of data points and 'n' is the number of independent
            variables.
        lb : jnp.ndarray
            Lower bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables.
        ub : jnp.ndarray
            Upper bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables.
        ftol : float
            Tolerance for termination by the change of the cost function.
        xtol : float
            Tolerance for termination by the change of the independent variables.
        gtol : float
            Tolerance for termination by the norm of the gradient.
        max_nfev : int
            Maximum number of function evaluations.
        f_scale : float
            Cost function scalar
        x_scale : jnp.ndarray
            Scaling factors for independent variables.
        loss_function : callable, optional
            Loss function. If None, the standard least-squares problem is
            solved.
        tr_options : dict
            Options for the trust-region algorithm.
        verbose : int
            Level of algorithm's verbosity:

                * 0 (default) : work silently.
                * 1 : display a termination report.

        timeit : bool, optional
            If True, the time for each step is measured if the unbounded
            version is being ran. Default is False.
        """
        # bounded or unbounded version
        if np.all(lb == -np.inf) and np.all(ub == np.inf):
            # unbounded version as timed and untimed version
            if not timeit:
                return self.trf_no_bounds(fun, xdata, ydata, jac, data_mask, 
                                          transform, x0, f0, J0,
                lb, ub, ftol, xtol, gtol, max_nfev, f_scale, x_scale, 
                loss_function, tr_options, verbose)
            else:
                return self.trf_no_bounds_timed(fun, xdata, ydata, jac, data_mask, transform, x0, 
                f0, J0, lb, ub, ftol, xtol, gtol, max_nfev, f_scale, x_scale, 
                loss_function, tr_options, verbose)
        else:
            return self.trf_bounds(fun, xdata, ydata, jac, data_mask, transform, x0, 
                    f0, J0, lb, ub, ftol, xtol, gtol, max_nfev, f_scale, x_scale, 
                    loss_function, tr_options, verbose)
        

    def trf_no_bounds(self, 
                      fun: Callable, 
                      xdata: Union[jnp.ndarray, Tuple[jnp.ndarray]], 
                      ydata: jnp.ndarray, 
                      jac: Callable, 
                      data_mask: jnp.ndarray, 
                      transform: jnp.ndarray, 
                      x0: np.ndarray, 
                      f: jnp.ndarray, 
                      J: jnp.ndarray, 
                      lb: np.ndarray, 
                      ub: np.ndarray,    
                      ftol: float, 
                      xtol: float, 
                      gtol: float, 
                      max_nfev: int,
                      f_scale: float, 
                      x_scale: np.ndarray, 
                      loss_function: Union[None, Callable], 
                      tr_options: Dict, 
                      verbose: int
                      ) -> Dict:
        """Unbounded version of the trust-region reflective algorithm.

        Parameters
        ----------
        fun : callable
            The residual function
        xdata : array_like or tuple of array_like
            The independent variable where the data is measured. If `xdata` is a
            tuple, then the input arguments to `fun` are assumed to be
            ``(xdata[0], xdata[1], ...)``.
        ydata : jnp.ndarray
            The dependent data
        jac : callable
            The Jacobian of `fun`.
        data_mask : jnp.ndarray
            The mask for the data.
        transform : jnp.ndarray
            The uncertainty transform for the data.
        x0 : jnp.ndarray
            Initial guess. Array of real elements of size (n,), where 'n' is the
            number of independent variables.
        f0 : jnp.ndarray
            Initial residuals. Array of real elements of size (m,), where 'm' is
            the number of data points.
        J0 : jnp.ndarray
            Initial Jacobian. Array of real elements of size (m, n), where 'm' is
            the number of data points and 'n' is the number of independent
            variables.
        lb : jnp.ndarray
            Lower bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables. 
        ub : jnp.ndarray
            Upper bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables.
        ftol : float
            Tolerance for termination by the change of the cost function.
        xtol : float
            Tolerance for termination by the change of the independent variables.
        gtol : float
            Tolerance for termination by the norm of the gradient.
        max_nfev : int
            Maximum number of function evaluations.
        f_scale : float
            Cost function scalar
        x_scale : jnp.ndarray
            Scaling factors for independent variables.
        loss_function : callable, optional
            Loss function. If None, the standard least-squares problem is
            solved.
        tr_options : dict
            Options for the trust-region algorithm.
        verbose : int
            Level of algorithm's verbosity: 

                * 0 (default) : work silently.
                * 1 : display a termination report.

        Returns
        -------
        result : OptimizeResult
            The optimization result represented as a ``OptimizeResult`` object.
            Important attributes are: ``x`` the solution array, ``success`` a
            Boolean flag indicating if the optimizer exited successfully and
            ``message`` which describes the cause of the termination. See
            `OptimizeResult` for a description of other attributes.

        Notes
        -----
        The algorithm is described in [13]_.

        """

        x = x0.copy()
        f_true = f
        nfev = 1
        njev = 1
        m, n = J.shape
    
        if loss_function is not None:
            rho = loss_function(f, f_scale)
            cost_jnp = self.calculate_cost(rho, data_mask)
            J, f = self.cJIT.scale_for_robust_loss_function(J, f, rho)
        else:      
            cost_jnp = self.default_loss_func(f)
        cost = np.array(cost_jnp)

        g_jnp = self.compute_grad(J, f)
        g = np.array(g_jnp)
        jac_scale = isinstance(x_scale, str) and x_scale == 'jac'
        if jac_scale:
            scale, scale_inv = self.cJIT.compute_jac_scale(J)
        else:
            scale, scale_inv = x_scale, 1 / x_scale
    
        Delta = norm(x0 * scale_inv)
        if Delta == 0:
            Delta = 1.0
    
        if max_nfev is None:
            max_nfev = x0.size * 100
    
        alpha = 0.0  # "Levenberg-Marquardt" parameter
    
        termination_status = None
        iteration = 0
        step_norm = None
        actual_reduction = None
    
        if verbose == 2:
            print_header_nonlinear()
    
        while True:
            g_norm = norm(g, ord=np.inf)
            if g_norm < gtol:
                termination_status = 1
                  
            if verbose == 2:
                print_iteration_nonlinear(iteration, nfev, cost, actual_reduction,
                                          step_norm, g_norm)
    
            if termination_status is not None or nfev == max_nfev:
                break
    
            d = scale
            d_jnp = jnp.array(scale)
            g_h_jnp = self.compute_grad_hat(g_jnp, d_jnp)
            svd_output = self.svd_no_bounds(J, d_jnp, f)

            J_h = svd_output[0]
            s, V, uf = [np.array(val) for val in svd_output[2:]]

            actual_reduction = -1
            while actual_reduction <= 0 and nfev < max_nfev:
                step_h, alpha, n_iter = solve_lsq_trust_region(
                    n, m, uf, s, V, Delta, initial_alpha=alpha)
    
                predicted_reduction_jnp = -self.cJIT.evaluate_quadratic(J_h, 
                                                                        g_h_jnp, 
                                                                        step_h)
                predicted_reduction = np.array(predicted_reduction_jnp)

                step = d * step_h
                x_new = x + step
                
                f_new = fun(x_new, xdata, ydata, data_mask, transform)
                
                nfev += 1
    
                step_h_norm = norm(step_h)
    
                if not self.check_isfinite(f_new):
                    Delta = 0.25 * step_h_norm
                    continue
    
                if loss_function is not None:
                    cost_new_jnp = loss_function(f_new, f_scale, data_mask, 
                                                 cost_only=True)
                else:
                    cost_new_jnp = self.default_loss_func(f_new)
                cost_new = np.array(cost_new_jnp)

                actual_reduction = cost - cost_new
 
                Delta_new, ratio = update_tr_radius(
                    Delta, actual_reduction, predicted_reduction,
                    step_h_norm, step_h_norm > 0.95 * Delta)
    
                step_norm = norm(step)
                termination_status = check_termination(
                    actual_reduction, cost, step_norm, norm(x), ratio, ftol, xtol)

                if termination_status is not None:
                    break
    
                alpha *= Delta / Delta_new
                Delta = Delta_new
    
            if actual_reduction > 0:
                x = x_new
                f = f_new
                f_true = f
                cost = cost_new
                J = jac(x, xdata, ydata, data_mask, transform)
                njev += 1
                
                if loss_function is not None:
                    rho = loss_function(f, f_scale)
                    J, f = self.cJIT.scale_for_robust_loss_function(J, f, rho)
                    
                g_jnp = self.compute_grad(J, f)
                g = np.array(g_jnp)
                if jac_scale:
                    scale, scale_inv = self.cJIT.compute_jac_scale(J, scale_inv)
            else:
                step_norm = 0
                actual_reduction = 0
    
            iteration += 1
    
        if termination_status is None:
            termination_status = 0
            
        active_mask = np.zeros_like(x)
        return OptimizeResult(
            x=x, cost=cost, fun=f_true, jac=J, grad=g, optimality=g_norm,
            active_mask=active_mask, nfev=nfev, njev=njev,
            status=termination_status, all_times={})


    def trf_bounds(self, 
                   fun: Callable, 
                    xdata: Union[jnp.ndarray, Tuple[jnp.ndarray]], 
                   ydata: jnp.ndarray, 
                   jac: Callable, 
                   data_mask: jnp.ndarray, 
                   transform: jnp.ndarray, 
                   x0: np.ndarray, 
                   f: jnp.ndarray, 
                   J: jnp.ndarray, 
                   lb: np.ndarray, 
                   ub: np.ndarray,    
                   ftol: float, 
                   xtol: float, 
                   gtol: float, 
                   max_nfev: int,
                   f_scale: float, 
                   x_scale: np.ndarray, 
                   loss_function: Union[None, Callable], 
                   tr_options: Dict, 
                   verbose: int
                   ) -> Dict:

        """Bounded version of the trust-region reflective algorithm.

        Parameters
        ----------
        fun : callable
            The residual function
        xdata : array_like or tuple of array_like
            The independent variable where the data is measured. If `xdata` is a
            tuple, then the input arguments to `fun` are assumed to be
            ``(xdata[0], xdata[1], ...)``.
        ydata : jnp.ndarray
            The dependent data
        jac : callable
            The Jacobian of `fun`.
        data_mask : jnp.ndarray
            The mask for the data.
        transform : jnp.ndarray
            The uncertainty transform for the data.
        x0 : jnp.ndarray
            Initial guess. Array of real elements of size (n,), where 'n' is the
            number of independent variables.
        f0 : jnp.ndarray
            Initial residuals. Array of real elements of size (m,), where 'm' is
            the number of data points.
        J0 : jnp.ndarray
            Initial Jacobian. Array of real elements of size (m, n), where 'm' is
            the number of data points and 'n' is the number of independent
            variables.
        lb : jnp.ndarray
            Lower bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables. 
        ub : jnp.ndarray
            Upper bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables.
        ftol : float
            Tolerance for termination by the change of the cost function.
        xtol : float
            Tolerance for termination by the change of the independent variables.
        gtol : float
            Tolerance for termination by the norm of the gradient.
        max_nfev : int
            Maximum number of function evaluations.
        f_scale : float
            Cost function scalar
        x_scale : jnp.ndarray
            Scaling factors for independent variables.
        loss_function : callable, optional
            Loss function. If None, the standard least-squares problem is
            solved.
        tr_options : dict
            Options for the trust-region algorithm.
        verbose : int
            Level of algorithm's verbosity: 

                * 0 (default) : work silently.
                * 1 : display a termination report.

        Returns
        -------
        result : OptimizeResult
            The optimization result represented as a ``OptimizeResult`` object.
            Important attributes are: ``x`` the solution array, ``success`` a
            Boolean flag indicating if the optimizer exited successfully and
            ``message`` which describes the cause of the termination. See
            `OptimizeResult` for a description of other attributes.

        Notes
        -----
        The algorithm is described in [13]_.

        References
        ----------
        .. [13] J. J. More, "The Levenberg-Marquardt Algorithm: Implementation and
                Theory," in Numerical Analysis, ed. G. A. Watson (1978), pp. 105-116.
                DOI: 10.1017/CBO9780511819595.006
        .. [2] T. F. Coleman and Y. Li, “An interior trust region approach for 
                nonlinear minimization subject to bounds,” SIAM Journal on 
                Optimization, vol. 6, no. 2, pp. 418–445, 1996.
        """
                
        x = x0.copy()
        f_true = f
        nfev = 1
        njev = 1
        m, n = J.shape
        
        if loss_function is not None:
            rho = loss_function(f, f_scale)
            cost_jnp = self.calculate_cost(rho, data_mask)
            J, f = self.cJIT.scale_for_robust_loss_function(J, f, rho)
        else:
            cost_jnp = self.default_loss_func(f)
            
        cost = np.array(cost_jnp)

        g_jnp = self.compute_grad(J, f)
        g = np.array(g_jnp)
        
        jac_scale = isinstance(x_scale, str) and x_scale == 'jac'
        if jac_scale:
            scale, scale_inv = self.cJIT.compute_jac_scale(J)
        else:
            scale, scale_inv = x_scale, 1 / x_scale
        

        v, dv = CL_scaling_vector(x, g, lb, ub)
    
        v[dv != 0] *= scale_inv[dv != 0]
        Delta = norm(x0 * scale_inv / v**0.5)
        if Delta == 0:
            Delta = 1.0
    
        g_norm = norm(g * v, ord=np.inf)
        
        if max_nfev is None:
            max_nfev = x0.size * 100
    
        alpha = 0.0  # "Levenberg-Marquardt" parameter
    
        termination_status = None
        iteration = 0
        step_norm = None
        actual_reduction = None
    
        if verbose == 2:
            print_header_nonlinear()
        
        f_zeros = jnp.zeros([n])
        while True:

            v, dv = CL_scaling_vector(x, g, lb, ub)
    
            g_norm = norm(g * v, ord=np.inf)
            if g_norm < gtol:
                termination_status = 1
    
            # if verbose == 2:
            #     print_iteration_nonlinear(iteration, nfev, cost, actual_reduction,
            #                               step_norm, g_norm)
    
            if termination_status is not None or nfev == max_nfev:
                break
    
            # Now compute variables in "hat" space. Here, we also account for
            # scaling introduced by `x_scale` parameter. This part is a bit tricky,
            # you have to write down the formulas and see how the trust-region
            # problem is formulated when the two types of scaling are applied.
            # The idea is that first we apply `x_scale` and then apply Coleman-Li
            # approach in the new variables.
    
            # v is recomputed in the variables after applying `x_scale`, note that
            # components which were identically 1 not affected.
            v[dv != 0] *= scale_inv[dv != 0]
    
            # Here, we apply two types of scaling.
            d = v**0.5 * scale
    
            # C = diag(g * scale) Jv
            diag_h = g * dv * scale
    
            # After all this has been done, we continue normally.
    
            # "hat" gradient.
            g_h = d * g
            J_diag = jnp.diag(diag_h**0.5)
            d_jnp = jnp.array(d)

            output = self.svd_bounds(f, J, d_jnp, J_diag, f_zeros)
            J_h = output[0]
            s, V, uf = [np.array(val) for val in output[2:]]

 
            # theta controls step back step ratio from the bounds.
            theta = max(0.995, 1 - g_norm)
    
            actual_reduction = -1
            while actual_reduction <= 0 and nfev < max_nfev:
                p_h, alpha, n_iter = solve_lsq_trust_region(
                    n, m, uf, s, V, Delta, initial_alpha=alpha)
                
                p = d * p_h  # Trust-region solution in the original space.
                step, step_h, predicted_reduction = self.select_step(
                    x, J_h, diag_h, g_h, p, p_h, d, Delta, lb, ub, theta)
                
                x_new = make_strictly_feasible(x + step.copy(), lb, ub, rstep=0)
                f_new = fun(x_new, xdata, ydata, data_mask, transform)

                nfev += 1
    
                step_h_norm = norm(step_h)
                if not self.check_isfinite(f_new):
                    Delta = 0.25 * step_h_norm
                    continue

                if loss_function is not None:
                    cost_new_jnp = loss_function(f_new, f_scale, data_mask, 
                                                 cost_only=True)
                else:
                    cost_new_jnp = self.default_loss_func(f_new)
                cost_new = np.array(cost_new_jnp)

                actual_reduction = cost - cost_new
                Delta_new, ratio = update_tr_radius(
                    Delta, actual_reduction, predicted_reduction,
                    step_h_norm, step_h_norm > 0.95 * Delta)

    
                step_norm = norm(step)
                termination_status = check_termination(
                    actual_reduction, cost, step_norm, norm(x), ratio, ftol, xtol)
                if termination_status is not None:
                    break
    
                alpha *= Delta / Delta_new
                Delta = Delta_new



            if actual_reduction > 0:
                x = x_new
                f = f_new
                f_true = f

                cost = cost_new
    
                J = jac(x, xdata, ydata, data_mask, transform)

                njev += 1
                
                if loss_function is not None:
                    rho = loss_function(f, f_scale)
                    J, f = self.cJIT.scale_for_robust_loss_function(J, f, rho)

                g_jnp = self.compute_grad(J, f)
                if jac_scale:
                    scale, scale_inv = self.cJIT.compute_jac_scale(J, scale_inv)
                g = np.array(g_jnp)
            else:
                step_norm = 0
                actual_reduction = 0
    
            iteration += 1
    
        if termination_status is None:
            termination_status = 0
                
        active_mask = find_active_constraints(x, lb, ub, rtol=xtol)
        return OptimizeResult(
            x=x, cost=cost, fun=f_true, jac=J, grad=g, optimality=g_norm,
            active_mask=active_mask, nfev=nfev, njev=njev,
            status=termination_status)
    
    def select_step(self, 
                    x: np.ndarray, 
                    J_h: jnp.ndarray, 
                    diag_h: jnp.ndarray, 
                    g_h: jnp.ndarray, 
                    p: np.ndarray, 
                    p_h: np.ndarray, 
                    d: np.ndarray, 
                    Delta: float, 
                    lb: np.ndarray, 
                    ub: np.ndarray, 
                    theta: float
                    ):
        """Select the best step according to Trust Region Reflective algorithm.
        
        Parameters
        ----------
        x : np.ndarray
            Current set parameter vector.
        J_h : jnp.ndarray
            Jacobian matrix in the scaled 'hat' space.
        diag_h : jnp.ndarray
            Diagonal of the scaled matrix C = diag(g * scale) Jv?
        g_h : jnp.ndarray
            Gradient vector in the scaled 'hat' space.
        p : np.ndarray
            Trust-region step in the original space.
        p_h : np.ndarray
            Trust-region step in the scaled 'hat' space.
        d : np.ndarray
            Scaling vector.
        Delta : float
            Trust-region radius.
        lb : np.ndarray
            Lower bounds on variables.
        ub : np.ndarray
            Upper bounds on variables.
        theta : float
            Controls step back step ratio from the bounds.

        Returns
        -------
        step : np.ndarray   
            Step in the original space.
        step_h : np.ndarray
            Step in the scaled 'hat' space.
        predicted_reduction : float
            Predicted reduction in the cost function.        
        """
        if in_bounds(x + p, lb, ub):
            p_value = self.cJIT.evaluate_quadratic(J_h, g_h, p_h, diag=diag_h)
            return p, p_h, -p_value
    
        p_stride, hits = step_size_to_bound(x, p, lb, ub)
    
        # Compute the reflected direction.
        r_h = np.copy(p_h)
        r_h[hits.astype(bool)] *= -1
        r = d * r_h
    
        # Restrict trust-region step, such that it hits the bound.
        p *= p_stride
        p_h *= p_stride
        x_on_bound = x + p
    
        # Reflected direction will cross first either feasible region or trust
        # region boundary.
        _, to_tr = intersect_trust_region(p_h, r_h, Delta)
        to_bound, _ = step_size_to_bound(x_on_bound, r, lb, ub)
    
        # Find lower and upper bounds on a step size along the reflected
        # direction, considering the strict feasibility requirement. There is no
        # single correct way to do that, the chosen approach seems to work best
        # on test problems.
        r_stride = min(to_bound, to_tr)
        if r_stride > 0:
            r_stride_l = (1 - theta) * p_stride / r_stride
            if r_stride == to_bound:
                r_stride_u = theta * to_bound
            else:
                r_stride_u = to_tr
        else:
            r_stride_l = 0
            r_stride_u = -1
    
        # Check if reflection step is available.
        if r_stride_l <= r_stride_u:
            a, b, c = self.cJIT.build_quadratic_1d(J_h, g_h, r_h, s0=p_h,
                                                   diag=diag_h)

            r_stride, r_value = minimize_quadratic_1d(
                a, b, r_stride_l, r_stride_u, c=c)
            r_h *= r_stride
            r_h += p_h
            r = r_h * d
        else:
            r_value = np.inf
    
        # Now correct p_h to make it strictly interior.
        p *= theta
        p_h *= theta
        p_value = self.cJIT.evaluate_quadratic(J_h, g_h, p_h, diag=diag_h)
    
        ag_h = -g_h
        ag = d * ag_h
    
        to_tr = Delta / norm(ag_h)
        to_bound, _ = step_size_to_bound(x, ag, lb, ub)
        if to_bound < to_tr:
            ag_stride = theta * to_bound
        else:
            ag_stride = to_tr
    
        a, b = self.cJIT.build_quadratic_1d(J_h, g_h, ag_h, diag=diag_h)
        ag_stride, ag_value = minimize_quadratic_1d(a, b, 0, ag_stride)
        ag_h *= ag_stride
        ag *= ag_stride
    
        if p_value < r_value and p_value < ag_value:
            return p, p_h, -p_value
        elif r_value < p_value and r_value < ag_value:
            return r, r_h, -r_value
        else:
            return ag, ag_h, -ag_value
        
        
    def trf_no_bounds_timed(self, 
                      fun: Callable, 
                      xdata: Union[jnp.ndarray, Tuple[jnp.ndarray]], 
                      ydata: jnp.ndarray, 
                      jac: Callable, 
                      data_mask: jnp.ndarray, 
                      transform: jnp.ndarray, 
                      x0: np.ndarray, 
                      f: jnp.ndarray, 
                      J: jnp.ndarray, 
                      lb: np.ndarray, 
                      ub: np.ndarray,    
                      ftol: float, 
                      xtol: float, 
                      gtol: float, 
                      max_nfev: int,
                      f_scale: float, 
                      x_scale: np.ndarray, 
                      loss_function: Union[None, Callable], 
                      tr_options: Dict, 
                      verbose: int
                      ) -> Dict:
        """Trust Region Reflective algorithm with no bounds and all the 
        operations performed on JAX and the GPU are timed. We need a separate
        function for this because to time each operation we need a 
        block_until_ready() function which makes the main Python thread wait
        until the GPU has finished the operation. However, for the main
        algorithm we don't want to wait for the GPU to finish each operation
        because it would slow down the algorithm. Thus, this is just used for
        analysis of the algorithm.

        Parameters
        ----------
        fun : callable
            The residual function
        xdata : array_like or tuple of array_like
            The independent variable where the data is measured. If `xdata` is a
            tuple, then the input arguments to `fun` are assumed to be
            ``(xdata[0], xdata[1], ...)``.
        ydata : jnp.ndarray
            The dependent data
        jac : callable
            The Jacobian of `fun`.
        data_mask : jnp.ndarray
            The mask for the data.
        transform : jnp.ndarray
            The uncertainty transform for the data.
        x0 : jnp.ndarray
            Initial guess. Array of real elements of size (n,), where 'n' is the
            number of independent variables.
        f0 : jnp.ndarray
            Initial residuals. Array of real elements of size (m,), where 'm' is
            the number of data points.
        J0 : jnp.ndarray
            Initial Jacobian. Array of real elements of size (m, n), where 'm' is
            the number of data points and 'n' is the number of independent
            variables.
        lb : jnp.ndarray
            Lower bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables. 
        ub : jnp.ndarray
            Upper bounds on independent variables. Array of real elements of size
            (n,), where 'n' is the number of independent variables.
        ftol : float
            Tolerance for termination by the change of the cost function.
        xtol : float
            Tolerance for termination by the change of the independent variables.
        gtol : float
            Tolerance for termination by the norm of the gradient.
        max_nfev : int
            Maximum number of function evaluations.
        f_scale : float
            Cost function scalar
        x_scale : jnp.ndarray
            Scaling factors for independent variables.
        loss_function : callable, optional
            Loss function. If None, the standard least-squares problem is
            solved.
        tr_options : dict
            Options for the trust-region algorithm.
        verbose : int
            Level of algorithm's verbosity: 

                * 0 (default) : work silently.
                * 1 : display a termination report.

        Returns
        -------
        result : OptimizeResult
            The optimization result represented as a ``OptimizeResult`` object.
            Important attributes are: ``x`` the solution array, ``success`` a
            Boolean flag indicating if the optimizer exited successfully and
            ``message`` which describes the cause of the termination. See
            `OptimizeResult` for a description of other attributes.

        Notes
        -----
        The algorithm is described in [13]_.
        """
      
        ftimes = []
        jtimes = []
        svd_times = []
        ctimes = []
        gtimes = []
        gtimes2 = []
        ptimes = []

        svd_ctimes = []
        g_ctimes = []
        c_ctimes = []
        p_ctimes = []
     
        x = x0.copy()
        
        st = time.time()
        f = fun(x, xdata, ydata, data_mask, transform).block_until_ready()
        ftimes.append(time.time() - st)
        f_true = f
        nfev = 1
        
        st = time.time()
        J = jac(x, xdata, ydata, data_mask, transform).block_until_ready()
        jtimes.append(time.time() - st)

        
        njev = 1
        m, n = J.shape
    
        if loss_function is not None:
            rho = loss_function(f, f_scale)
            cost_jnp = self.calculate_cost(rho, data_mask)
            J, f = self.cJIT.scale_for_robust_loss_function(J, f, rho)
        else:      
            st1 = time.time()
            cost_jnp = self.default_loss_func(f).block_until_ready()
            st2 = time.time()
        cost = np.array(cost_jnp)
        st3 = time.time()

        ctimes.append(st2 - st1)
        c_ctimes.append(st3 - st2)
        
        st1 = time.time()
        g_jnp = self.compute_grad(J, f).block_until_ready()
        st2 = time.time()
        g = np.array(g_jnp)
        st3 = time.time()


        gtimes.append(st2 - st1)
        g_ctimes.append(st3 - st2)
        
        jac_scale = isinstance(x_scale, str) and x_scale == 'jac'
        if jac_scale:
            scale, scale_inv = self.cJIT.compute_jac_scale(J)
        else:
            scale, scale_inv = x_scale, 1 / x_scale
    
        Delta = norm(x0 * scale_inv)
        if Delta == 0:
            Delta = 1.0
    
        if max_nfev is None:
            max_nfev = x0.size * 100
    
        alpha = 0.0  # "Levenberg-Marquardt" parameter
    
        termination_status = None
        iteration = 0
        step_norm = None
        actual_reduction = None
    
        if verbose == 2:
            print_header_nonlinear()
    
        while True:
            g_norm = norm(g, ord=np.inf)
            if g_norm < gtol:
                termination_status = 1
                  
            # if verbose == 2:
            #     print_iteration_nonlinear(iteration, nfev, cost, actual_reduction,
            #                               step_norm, g_norm)
    
            if termination_status is not None or nfev == max_nfev:
                break
    
            d = scale
            d_jnp = jnp.array(scale)

            # g_h = d * g
            g_h_jnp = self.compute_grad_hat(g_jnp, d_jnp)
            
            st = time.time()
            svd_output = self.svd_no_bounds(J, d_jnp, f)
            tree_flatten(svd_output)[0][0].block_until_ready()
            svd_times.append(time.time() - st)
            J_h = svd_output[0]
            
            st = time.time()
            s, V, uf = [np.array(val) for val in svd_output[2:]]
            svd_ctimes.append(time.time() - st)

            
            actual_reduction = -1
            while actual_reduction <= 0 and nfev < max_nfev:
                step_h, alpha, n_iter = solve_lsq_trust_region(
                    n, m, uf, s, V, Delta, initial_alpha=alpha)
    
                st1 = time.time()
                predicted_reduction_jnp = -self.cJIT.evaluate_quadratic(J_h, g_h_jnp, 
                                                                    step_h).block_until_ready()
                st2 = time.time()
                predicted_reduction = np.array(predicted_reduction_jnp)
                st3 = time.time()
                ptimes.append(st2 - st1)
                p_ctimes.append(st3 - st2)

                step = d * step_h
                x_new = x + step
                
                st = time.time()
                f_new = fun(x_new, xdata, ydata, data_mask, transform).block_until_ready()
                ftimes.append(time.time() - st)
                
                nfev += 1
    
                step_h_norm = norm(step_h)
    
                if not self.check_isfinite(f_new):
                    Delta = 0.25 * step_h_norm
                    continue
    
                if loss_function is not None:
                    cost_new_jnp = loss_function(f_new, f_scale, data_mask, 
                                                 cost_only=True)
                else:
                    st1 = time.time()
                    cost_new_jnp = self.default_loss_func(f_new).block_until_ready()
                    st2 = time.time()
                    cost_new = np.array(cost_new_jnp)
                    st3 = time.time()

                    ctimes.append(st2 - st1)
                    c_ctimes.append(st3 - st2)

                actual_reduction = cost - cost_new
 
                Delta_new, ratio = update_tr_radius(
                    Delta, actual_reduction, predicted_reduction,
                    step_h_norm, step_h_norm > 0.95 * Delta)
    
                step_norm = norm(step)
                termination_status = check_termination(
                    actual_reduction, cost, step_norm, norm(x), ratio, ftol, xtol)

                if termination_status is not None:
                    break
    
                alpha *= Delta / Delta_new
                Delta = Delta_new
    
            if actual_reduction > 0:
                x = x_new
    
                f = f_new
                f_true = f
    
                cost = cost_new
                
                st = time.time()
                J = jac(x, xdata, ydata, data_mask, transform).block_until_ready()
                jtimes.append(time.time() - st)
                
                njev += 1
                
                if loss_function is not None:
                    rho = loss_function(f, f_scale)
                    J, f = self.cJIT.scale_for_robust_loss_function(J, f, rho)
                    
                    
                st1 = time.time()
                g_jnp = self.compute_grad(J, f).block_until_ready()
                st2 = time.time()
                g = np.array(g_jnp)
                st3 = time.time()

                gtimes.append(st2 - st1)
                g_ctimes.append(st3 - st2)
    
                if jac_scale:
                    scale, scale_inv = self.cJIT.compute_jac_scale(J, scale_inv)
                
            else:
                step_norm = 0
                actual_reduction = 0
    
            iteration += 1
    
        if termination_status is None:
            termination_status = 0
            
        active_mask = np.zeros_like(x)

        tlabels = ['ftimes', 'jtimes', 'svd_times', 'ctimes', 'gtimes', 'ptimes', 
                   'g_ctimes', 'c_ctimes', 'svd_ctimes', 'p_ctimes', 'gtimes2']
        all_times = [ftimes, jtimes, svd_times, ctimes, gtimes, ptimes, 
                     g_ctimes, c_ctimes, svd_ctimes, p_ctimes, gtimes2]
                     
        tdicts = dict(zip(tlabels, all_times))
        return OptimizeResult(
            x=x, cost=cost, fun=f_true, jac=J, grad=g, optimality=g_norm,
            active_mask=active_mask, nfev=nfev, njev=njev,
            status=termination_status, all_times=tdicts)


import warnings
import numpy as np
from numpy import (zeros, inf)
from inspect import signature
import time
from typing import Optional, Callable, Tuple, Union, List, Dict, Any

from jax.config import config
config.update("jax_enable_x64", True)

import jax.numpy as jnp
from jax import jit
from jax.scipy.linalg import svd as jax_svd
from jax.scipy.linalg import cholesky as jax_cholesky

from jaxfit._optimize import OptimizeWarning
from jaxfit.least_squares import prepare_bounds, LeastSquares
from jaxfit.common_scipy import EPS

__all__ = ['CurveFit', 'curve_fit']

def curve_fit(f, *args, **kwargs):
    jcf = CurveFit(f)
    popt, pcov, _, _, _ = jcf.curve_fit(*args, **kwargs)
    return popt, pcov


def _initialize_feasible(lb: np.ndarray, ub: np.ndarray) -> np.ndarray:
    """Initialize feasible parameters for optimization.

    This function initializes feasible parameters for optimization based on the 
    lower and upper bounds of the variables. If both bounds are finite, the 
    feasible parameters are set to the midpoint between the bounds. If only the 
    lower bound is finite, the feasible parameters are set to the lower bound 
    plus 1. If only the upper bound is finite, the feasible parameters are set 
    to the upper bound minus 1. If neither bound is finite, the feasible 
    parameters are set to 1.

    Parameters
    ----------
    lb : np.ndarray
        The lower bounds of the variables.
    ub : np.ndarray
        The upper bounds of the variables.

    Returns
    -------
    np.ndarray
        The initialized feasible parameters.
    """

    p0 = np.ones_like(lb)
    lb_finite = np.isfinite(lb)
    ub_finite = np.isfinite(ub)

    mask = lb_finite & ub_finite
    p0[mask] = 0.5 * (lb[mask] + ub[mask])

    mask = lb_finite & ~ub_finite
    p0[mask] = lb[mask] + 1

    mask = ~lb_finite & ub_finite
    p0[mask] = ub[mask] - 1

    return p0


class CurveFit():
    
    def __init__(self, flength: Optional[float] = None):
        """CurveFit class for fitting

        Parameters
        ----------
        flength : float, optional
            fixed data length for fits, JAXFit pads input data to this length 
            to avoid retracing.
        """
        self.flength = flength
        self.create_sigma_transform_funcs()
        self.create_covariance_svd()
        self.ls = LeastSquares()


    def update_flength(self, flength: float):
        """Set the fixed input data length.

        Parameters
        ----------
        flength : float
            The fixed input data length.
        """
        self.flength = flength
                
        
    def create_sigma_transform_funcs(self):
        """Create JIT-compiled sigma transform functions.

        This function creates two JIT-compiled functions: `sigma_transform1d` and 
        `sigma_transform2d`, which are used to compute the sigma transform for 1D 
        and 2D data, respectively. The functions are stored as attributes of the 
        object on which the method is called.
        """

        @jit
        def sigma_transform1d(sigma: jnp.ndarray, 
                      data_mask: jnp.ndarray
                     ) -> jnp.ndarray:
            """Compute the sigma transform for 1D data.

            Parameters
            ----------
            sigma : jnp.ndarray
                The standard deviation of the data.
            data_mask : jnp.ndarray
                A binary mask indicating which data points to use in the fit.

            Returns
            -------
            jnp.ndarray
                The sigma transform for the data.
            """
            transform = 1.0 / sigma
            return transform

        @jit
        def sigma_transform2d(sigma: jnp.ndarray, 
                              data_mask: jnp.ndarray
                              ) -> jnp.ndarray:
            """Compute the sigma transform for 2D data.

            Parameters
            ----------
            sigma : jnp.ndarray
                The standard deviation of the data.
            data_mask : jnp.ndarray
                A binary mask indicating which data points to use in the fit.

            Returns
            -------
            jnp.ndarray
                The sigma transform for the data.
            """
            sigma = jnp.asarray(sigma)
            transform = jax_cholesky(sigma, lower=True)
            return transform

        self.sigma_transform1d = sigma_transform1d
        self.sigma_transform2d = sigma_transform2d
        """For fixed input arrays we need to pad the actual data to match the 
        fixed input array size"""

    def create_covariance_svd(self): 
        """Create JIT-compiled SVD function for covariance computation."""
        @jit
        def covariance_svd(jac):
            _, s, VT = jax_svd(jac, full_matrices=False)
            return s, VT
        self.covariance_svd = covariance_svd
        
        
    def pad_fit_data(self, xdata: np.ndarray, 
                     ydata: np.ndarray, 
                     xdims: int, 
                     len_diff: int
                     ) -> Tuple[np.ndarray, np.ndarray]:
        """Pad fit data to match the fixed input data length.

        This function pads the input data arrays with small values to match the 
        fixed input data length to avoid JAX retracing the JITted functions.
        The padding is added along the second dimension of the `xdata` array 
        if it's multidimensional data otherwise along the first dimension. The 
        small values are chosen to be `EPS`, a global constant defined as a 
        very small positive value which avoids numerical issues.

        Parameters
        ----------
        xdata : np.ndarray
            The independent variables of the data.
        ydata : np.ndarray
            The dependent variables of the data.
        xdims : int
            The number of dimensions in the `xdata` array.
        len_diff : int
            The difference in length between the data arrays and the fixed input data length.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The padded `xdata` and `ydata` arrays.
        """
        
        if xdims > 1:
            xpad = EPS * np.ones([xdims, len_diff])
            xdata = np.concatenate([xdata, xpad], axis=1)
        else:
            xpad = EPS * np.ones([len_diff])
            xdata = np.concatenate([xdata, xpad])
        ypad = EPS * np.ones([len_diff])
        ydata = np.concatenate([ydata, ypad])
        return xdata, ydata

        
        
    def curve_fit(self, 
                  f: Callable, 
                  xdata: Union[np.ndarray, Tuple[np.ndarray]],
                  ydata: np.ndarray, 
                  p0: Optional[np.ndarray] = None, 
                  sigma: Optional[np.ndarray] = None, 
                  absolute_sigma: bool = False, 
                  check_finite: bool = True,
                  bounds: Tuple[np.ndarray, np.ndarray] = (-np.inf, np.inf), 
                  method: Optional[str] = None, 
                  jac: Optional[Callable] = None,
                  data_mask: Optional[np.ndarray] = None, 
                  timeit: bool = False, 
                  return_eval: bool = False,
                  **kwargs
                  ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Use non-linear least squares to fit a function, f, to data.
        Assumes ``ydata = f(xdata, *params) + eps``.

        Parameters
        ----------
        f : callable
            The model function, f(x, ...). It must take the independent
            variable as the first argument and the parameters to fit as
            separate remaining arguments.
        xdata : array_like or object
            The independent variable where the data is measured.
            Should usually be an M-length sequence or an (k,M)-shaped array for
            functions with k predictors, but can actually be any object.
        ydata : array_like
            The dependent data, a length M array - nominally ``f(xdata, ...)``.
        p0 : array_like, optional
            Initial guess for the parameters (length N). If None, then the
            initial values will all be 1 (if the number of parameters for the
            function can be determined using introspection, otherwise a
            ValueError is raised).
        sigma : None or M-length sequence or MxM array, optional
            Determines the uncertainty in `ydata`. If we define residuals as
            ``r = ydata - f(xdata, *popt)``, then the interpretation of `sigma`
            depends on its number of dimensions:
            - A 1-D `sigma` should contain values of standard deviations of
            errors in `ydata`. In this case, the optimized function is
            ``chisq = sum((r / sigma) ** 2)``.
            - A 2-D `sigma` should contain the covariance matrix of
            errors in `ydata`. In this case, the optimized function is
            ``chisq = r.T @ inv(sigma) @ r``.
            .. versionadded:: 0.19

            None (default) is equivalent of 1-D `sigma` filled with ones.
        absolute_sigma : bool, optional
            If True, `sigma` is used in an absolute sense and the estimated parameter
            covariance `pcov` reflects these absolute values.
            If False (default), only the relative magnitudes of the `sigma` values matter.
            The returned parameter covariance matrix `pcov` is based on scaling
            `sigma` by a constant factor. This constant is set by demanding that the
            reduced `chisq` for the optimal parameters `popt` when using the
            *scaled* `sigma` equals unity. In other words, `sigma` is scaled to
            match the sample variance of the residuals after the fit. Default is False.
            Mathematically,
            ``pcov(absolute_sigma=False) = pcov(absolute_sigma=True) * chisq(popt)/(M-N)``
        check_finite : bool, optional
            If True, check that the input arrays do not contain nans of infs,
            and raise a ValueError if they do. Setting this parameter to
            False may silently produce nonsensical results if the input arrays
            do contain nans. Default is True.
        bounds : 2-tuple of array_like, optional
            Lower and upper bounds on parameters. Defaults to no bounds.
            Each element of the tuple must be either an array with the length equal
            to the number of parameters, or a scalar (in which case the bound is
            taken to be the same for all parameters). Use ``np.inf`` with an
            appropriate sign to disable bounds on all or some parameters.
            .. versionadded:: 0.17
        method : {'trf'}, optional
            Method to use for optimization. See `least_squares` for more details.
            Currently only 'trf' is implemented.
            .. versionadded:: 0.17
        jac : callable, string or None, optional
            Function with signature ``jac(x, ...)`` which computes the Jacobian
            matrix of the model function with respect to parameters as a dense
            array_like structure. It will be scaled according to provided `sigma`.
            If None (default), the Jacobian will be determined using JAX's automatic
            differentiation (AD) capabilities. We recommend not using an analytical
            Jacobian, as it is usually faster to use AD.
        kwargs
            Keyword arguments passed to `leastsq` for ``method='lm'`` or
            `least_squares` otherwise.

        Returns
        -------
        popt : array
            Optimal values for the parameters so that the sum of the squared
            residuals of ``f(xdata, *popt) - ydata`` is minimized.
        pcov : 2-D array
            The estimated covariance of popt. The diagonals provide the variance
            of the parameter estimate. To compute one standard deviation errors
            on the parameters use ``perr = np.sqrt(np.diag(pcov))``.
            How the `sigma` parameter affects the estimated covariance
            depends on `absolute_sigma` argument, as described above.
            If the Jacobian matrix at the solution doesn't have a full rank, then
            'lm' method returns a matrix filled with ``np.inf``, on the other hand
            'trf'  and 'dogbox' methods use Moore-Penrose pseudoinverse to compute
            the covariance matrix.

        Raises
        ------
        ValueError
            if either `ydata` or `xdata` contain NaNs, or if incompatible options
            are used.
        RuntimeError
            if the least-squares minimization fails.
        OptimizeWarning
            if covariance of the parameters can not be estimated.
        See Also
        --------
        least_squares : Minimize the sum of squares of nonlinear functions.
        
        Notes
        -----
        Refer to the docstring of `least_squares` for more information.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> import jax.numpy as jnp
        >>> from jaxfit import CurveFit
        >>> def func(x, a, b, c):
        ...     return a * jnp.exp(-b * x) + c
        Define the data to be fit with some noise:
        >>> xdata = np.linspace(0, 4, 50)
        >>> y = func(xdata, 2.5, 1.3, 0.5)
        >>> rng = np.random.default_rng()
        >>> y_noise = 0.2 * rng.normal(size=xdata.size)
        >>> ydata = y + y_noise
        >>> plt.plot(xdata, ydata, 'b-', label='data')
        Fit for the parameters a, b, c of the function `func`:
        >>> cf = CurveFit()
        >>> popt, pcov = cf.curve_fit(func, xdata, ydata)
        >>> popt
        array([2.56274217, 1.37268521, 0.47427475])
        >>> plt.plot(xdata, func(xdata, *popt), 'r-',
        ...          label='fit: a=%5.3f, b=%5.3f, c=%5.3f' % tuple(popt))
        Constrain the optimization to the region of ``0 <= a <= 3``,
        ``0 <= b <= 1`` and ``0 <= c <= 0.5``:
        >>> cf = CurveFit()
        >>> popt, pcov = cf.curve_fit(func, xdata, ydata, bounds=(0, [3., 1., 0.5]))
        >>> popt
        array([2.43736712, 1.        , 0.34463856])
        >>> plt.plot(xdata, func(xdata, *popt), 'g--',
        ...          label='fit: a=%5.3f, b=%5.3f, c=%5.3f' % tuple(popt))
        >>> plt.xlabel('x')
        >>> plt.ylabel('y')
        >>> plt.legend()
        >>> plt.show()
        """
        
        if p0 is None:
            # determine number of parameters by inspecting the function
            sig = signature(f)
            args = sig.parameters
            if len(args) < 2:
                raise ValueError("Unable to determine number of fit parameters.")
            n = len(args) - 1
        else:
            p0 = np.atleast_1d(p0)
            n = p0.size
                    
        lb, ub = prepare_bounds(bounds, n)
        if p0 is None:
            p0 = _initialize_feasible(lb, ub)
            
        if method is None:
            method = 'trf'

        # NaNs cannot be handled
        if check_finite:
            ydata = np.asarray_chkfinite(ydata, float)
        else:
            ydata = np.asarray(ydata, float)
    
        if isinstance(xdata, (list, tuple, np.ndarray)):
            #should we be able to pass jax arrays
            # `xdata` is passed straight to the user-defined `f`, so allow
            # non-array_like `xdata`.
            if check_finite:
                xdata = np.asarray_chkfinite(xdata, float)
            else:
                xdata = np.asarray(xdata, float)
        else:
            raise ValueError('X needs arrays')
    
        if ydata.size == 0:
            raise ValueError("`ydata` must not be empty!")
            
        m = len(ydata)            
        xdims = xdata.ndim
        if xdims == 1:
            xlen = len(xdata)
        else:
            xlen = len(xdata[0])
        if xlen != m:
            print(xdata.shape, ydata.shape)
            raise ValueError('X and Y data lengths dont match')
        
        if data_mask is None:
            none_mask = True
        else:
            none_mask = False
            
        if self.flength is not None:
            len_diff = self.flength - m
            if data_mask is not None:
                if len(data_mask) != m:
                    raise ValueError('Data mask doesnt match data lengths.')
            else:
                data_mask = np.ones(m, dtype=bool)
                if len_diff > 0:
                    data_mask = np.concatenate([data_mask, 
                                                np.zeros(len_diff, dtype=bool)])
        else:
            len_diff = 0
            data_mask = np.ones(m, dtype=bool)
            
        
        if self.flength is not None:
            if len_diff >= 0:
                xdata, ydata = self.pad_fit_data(xdata, ydata, xdims, len_diff)
            else:
                print('Data length greater than fixed length. This means retracing will occur')
            
                   # Determine type of sigma
        if sigma is not None:   
            if not isinstance(sigma, np.ndarray):
                raise ValueError('Sigma must be numpy array.')
            # if 1-D, sigma are errors, define transform = 1/sigma
            ysize = ydata.size - len_diff
            if sigma.shape == (ysize, ):
                if len_diff > 0:
                    sigma = np.concatenate([sigma, np.ones([len_diff])])
                transform = self.sigma_transform1d(sigma, data_mask)
            # if 2-D, sigma is the covariance matrix,
            # define transform = L such that L L^T = C
            elif sigma.shape == (ysize, ysize):
                try:
                    if len_diff >= 0:
                        sigma_padded = np.identity(m + len_diff)
                        sigma_padded[:m,:m] = sigma
                        sigma = sigma_padded
                    # scipy.linalg.cholesky requires lower=True to return L L^T = A
                    transform = self.sigma_transform2d(sigma, data_mask)
                except:
                    raise ValueError("Probably:`sigma` must be positive definite.")
            else:
                print('sigma shape', sigma.shape)
                print('y shape', ydata.shape, ydata.size)
                print(len_diff)
                raise ValueError("`sigma` has incorrect shape.")
        else:
            transform = None
            

        if 'args' in kwargs:
            # The specification for the model function `f` does not support
            # additional arguments. Refer to the `curve_fit` docstring for
            # acceptable call signatures of `f`.
            raise ValueError("'args' is not a supported keyword argument.")
    
        if 'max_nfev' not in kwargs:
            kwargs['max_nfev'] = kwargs.pop('maxfev', None)
        
        st = time.time()
        if timeit:
            jnp_xdata = jnp.array(np.copy(xdata)).block_until_ready()
            jnp_ydata = jnp.array(np.copy(ydata)).block_until_ready()
        else:
            jnp_xdata = jnp.array(np.copy(xdata))
            jnp_ydata = jnp.array(np.copy(ydata))
        ctime = time.time() - st

        jnp_data_mask = jnp.array(data_mask, dtype=bool)
        res = self.ls.least_squares(f, p0, jac=jac, xdata=jnp_xdata, 
                                    ydata=jnp_ydata, data_mask=jnp_data_mask, 
                                    transform=transform, bounds=bounds, 
                                    method=method, timeit=timeit, **kwargs)

        if not res.success:
            raise RuntimeError("Optimal parameters not found: " + res.message)
        popt = res.x

        st = time.time()
        # ysize = len(res.fun)
        ysize = m
        cost = 2 * res.cost  # res.cost is half sum of squares!

        # Do Moore-Penrose inverse discarding zero singular values.
        # _, s, VT = svd(res.jac, full_matrices=False)
        outputs = self.covariance_svd(res.jac)
        s, VT = [np.array(output) for output in outputs]
        threshold = np.finfo(float).eps * max(res.jac.shape) * s[0]
        s = s[s > threshold]
        VT = VT[:s.size]
        pcov = np.dot(VT.T / s**2, VT)
        return_full = False
    
        warn_cov = False
        if pcov is None:
            # indeterminate covariance
            pcov = zeros((len(popt), len(popt)), dtype=float)
            pcov.fill(inf)
            warn_cov = True
        elif not absolute_sigma:
            if ysize > p0.size:
                s_sq = cost / (ysize - p0.size)
                pcov = pcov * s_sq
            else:
                pcov.fill(inf)
                warn_cov = True
    
        if warn_cov:
            warnings.warn('Covariance of the parameters could not be estimated',
                          category=OptimizeWarning)

        # self.res = res
        post_time = time.time() - st

        if return_eval:
            feval = f(jnp_xdata, *popt)
            feval = np.array(feval)
            if none_mask:
                # data_mask = np.ndarray.astype(data_mask, bool)
                return popt, pcov, feval[data_mask]
            else:
                return popt, pcov, feval
        else:
            #lower GPU memory usage
            res.pop('jac')
            res.pop('fun')

        if return_full:
            raise RuntimeError("Return full only works for LM")
            # return popt, pcov, infodict, errmsg, ier
        elif timeit:
            return popt, pcov, res, post_time, ctime
        else:
            return popt, pcov

            

    


"""
These are functions that were initially in the common.py file, but are have
large data operations and are therefore better suited to be compiled with
JAX.  They are compiled with JAX and then added to the CommonJIT class.
"""
import numpy as np
from jax.config import config
config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax import jit
import time
from typing import Tuple, List, Dict, Union, Optional, Callable

EPS = np.finfo(float).eps


class CommonJIT():
    
    def __init__(self):
        """Initialize the class and create the JAX/JIT functions that will be 
        compiled"""
        self.create_quadratic_funcs()
        self.create_js_dot()
        self.create_jac_sum()
        self.create_scale_for_robust_loss_function()
        
    
    def create_scale_for_robust_loss_function(self):
        """Create the scaling function for the loss functions"""

        @jit
        def scale_for_robust_loss_function(J: jnp.ndarray, 
                                           f: jnp.ndarray,
                                           rho: jnp.ndarray
                                           ) -> Tuple[jnp.ndarray, jnp.ndarray]:
            """Scale Jacobian and residuals for a robust loss function.
            Arrays are modified in place.

            Parameters
            ----------
            J : jnp.ndarray
                Jacobian matrix.
            f : jnp.ndarray
                Residuals.
            rho : jnp.ndarray
                Cost function evaluation.
            """
            J_scale = rho[1] + 2 * rho[2] * f**2
            mask = J_scale < EPS
            J_scale = jnp.where(mask, EPS, J_scale)
            J_scale = J_scale**0.5
            fscale = (rho[1] / J_scale)

            f = f * fscale
            J = J * J_scale[:, jnp.newaxis]
            return J, f
        self.scale_for_robust_loss_function = scale_for_robust_loss_function    
    
    
    def build_quadratic_1d(self, 
                           J: jnp.ndarray, 
                           g: jnp.ndarray, 
                           s: jnp.ndarray, 
                           diag: Optional[jnp.ndarray] = None, 
                           s0: Optional[jnp.ndarray] = None
                           ) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray], 
                                      Tuple[np.ndarray, np.ndarray]]:

        """Parameterize a multivariate quadratic function along a line.

        The resulting univariate quadratic function is given as follows:

            f(t) = 0.5 * (s0 + s*t).T * (J.T*J + diag) * (s0 + s*t) + g.T * (s0 + s*t)

        Parameters
        ----------
        J : ndarray, sparse matrix or LinearOperator, shape (m, n)
            Jacobian matrix, affects the quadratic term.
        g : ndarray, shape (n,)
            Gradient, defines the linear term.
        s : ndarray, shape (n,)
            Direction vector of a line.
        diag : None or ndarray with shape (n,), optional
            Addition diagonal part, affects the quadratic term.
            If None, assumed to be 0.
        s0 : None or ndarray with shape (n,), optional
            Initial point. If None, assumed to be 0.

        Returns
        -------
        a : float
            Coefficient for t**2.
        b : float
            Coefficient for t.
        c : float
            Free term. Returned only if `s0` is provided.
            
        """

        s_jnp = jnp.array(s)
        v_jnp = self.js_dot(J, s_jnp)
        v = v_jnp.copy()

        a = np.dot(v, v)
        if diag is not None:
            a += np.dot(s * diag, s)
        a *= 0.5
    
        b = np.dot(g, s)
    
        if s0 is not None:
            s0_jnp = jnp.array(s0)
            u_jnp = self.js0_dot(J, s0_jnp)
            u = u_jnp.copy()

            b += np.dot(u, v)
            c = 0.5 * np.dot(u, u) + np.dot(g, s0)
            if diag is not None:
                b += np.dot(s0 * diag, s)
                c += 0.5 * np.dot(s0 * diag, s0)
            return a, b, c
        else:
            return a, b


    def compute_jac_scale(self, J: jnp.ndarray, 
                          scale_inv_old: Optional[np.ndarray] = None
                          ) -> Tuple[np.ndarray, np.ndarray]:
        """Compute variables scale based on the Jacobian matrix.
        
        Parameters
        ----------
        J : jnp.ndarray
            Jacobian matrix.
        scale_inv_old : Optional[np.ndarray], optional
            Previous scale, by default None

        Returns
        -------
        scale : np.ndarray
            Scale for the variables.
        scale_inv : np.ndarray
            Inverse of the scale for the variables.
        """

        scale_inv_jnp = self.jac_sum_func(J)
        scale_inv = np.array(scale_inv_jnp)
    
        if scale_inv_old is None:
            scale_inv[scale_inv == 0] = 1
        else:
            scale_inv = np.maximum(scale_inv, scale_inv_old)

        return 1 / scale_inv, scale_inv
    

    def create_js_dot(self):
        """Create the functions for the dot product of the Jacobian and the
        search direction. We need two functions because s and s0 are different
        shapes which causes retracing of the function if we use the same
        function for both.
        """
        @jit
        def js_dot(J: jnp.ndarray, s: jnp.ndarray) -> jnp.ndarray:
            return J.dot(s)

        @jit
        def js0_dot(J: jnp.ndarray, s0: jnp.ndarray) -> jnp.ndarray:
            return J.dot(s0)

        self.js_dot = js_dot
        self.js0_dot = js0_dot

  
    def evaluate_quadratic(self, 
                           J: jnp.ndarray, 
                           g: jnp.ndarray, 
                           s_np: np.ndarray, 
                           diag: Optional[np.ndarray] = None
                           ) -> jnp.ndarray:
        """Compute values of a quadratic function arising in least squares.
        The function is 0.5 * s.T * (J.T * J + diag) * s + g.T * s.

        Parameters
        ----------
        J : ndarray, sparse matrix or LinearOperator, shape (m, n)
            Jacobian matrix, affects the quadratic term.
        g : ndarray, shape (n,)
            Gradient, defines the linear term.
        s : ndarray, shape (k, n) or (n,)
            Array containing steps as rows.
        diag : ndarray, shape (n,), optional
            Addition diagonal part, affects the quadratic term.
            If None, assumed to be 0.
        Returns
        -------
        values : ndarray with shape (k,) or float
            Values of the function. If `s` was 2-D, then ndarray is
            returned, otherwise, float is returned.
        """
        s = jnp.array(s_np) #comes in as np array
        
        if s.ndim == 1:
            if diag is None:
                return self.evaluate_quadratic1(J, g, s)
            else:
                return self.evaluate_quadratic_diagonal1(J, g, s, diag)
        else:
            if diag is None:
                return self.evaluate_quadratic2(J, g, s)
            else:
                return self.evaluate_quadratic_diagonal2(J, g, s, diag)
    
    
    def create_quadratic_funcs(self):
        
        @jit
        def evaluate_quadratic1(J, g, s):
            Js = J.dot(s)
            q = jnp.dot(Js, Js)
            l = jnp.dot(s, g)
            return 0.5 * q + l
        
        @jit
        def evaluate_quadratic_diagonal1(J, g, s, diag):
            Js = J.dot(s)
            q = jnp.dot(Js, Js) + jnp.dot(s * diag, s)
            l = jnp.dot(s, g)
            return 0.5 * q + l
        
        @jit
        def evaluate_quadratic2(J, g, s):
            Js = J.dot(s.T)
            q = jnp.sum(Js**2, axis=0)
            l = jnp.dot(s, g)
            return 0.5 * q + l
        
        @jit
        def evaluate_quadratic_diagonal2(J, g, s, diag):
            Js = J.dot(s.T)
            q = jnp.sum(Js**2, axis=0) + jnp.sum(diag * s**2, axis=1)
            l = jnp.dot(s, g)
            return 0.5 * q + l
        
        self.evaluate_quadratic1 = evaluate_quadratic1
        self.evaluate_quadratic_diagonal1 = evaluate_quadratic_diagonal1
        self.evaluate_quadratic2 = evaluate_quadratic2
        self.evaluate_quadratic_diagonal2 = evaluate_quadratic_diagonal2


    def create_jac_sum(self):
        """Create the function for the sum of the Jacobian squared and then
        taking the square root. This is used to compute the scale for the
        variables. Can potentially remove this.
        """
        @jit
        def jac_sum_func(J):
            return jnp.sum(J**2, axis=0)**0.5
        self.jac_sum_func = jac_sum_func


from jax.config import config
config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax import jit


class LossFunctionsJIT():

    def __init__(self):
        self.stack_rhos = self.create_stack_rhos()

        self.create_huber_funcs()
        self.create_soft_l1_funcs()
        self.create_cauchy_funcs()
        self.create_arctan_funcs()

        self.IMPLEMENTED_LOSSES = dict(linear=None, huber=self.huber,
                                       soft_l1=self.soft_l1,
                                       cauchy=self.cauchy,
                                       arctan=self.arctan)
        self.loss_funcs = self.construct_all_loss_functions()

        self.create_zscale()
        self.create_calculate_cost()
        self.create_scale_rhos()

    def create_stack_rhos(self):
        @jit
        def stack_rhos(rho0, rho1, rho2):
            return jnp.stack([rho0, rho1, rho2])

        return stack_rhos

    def get_empty_rhos(self, z):
        dlength = len(z)
        rho1 = jnp.zeros([dlength])
        rho2 = jnp.zeros([dlength])
        return rho1, rho2

    def create_huber_funcs(self):
        @jit
        def huber1(z):
            mask = z <= 1

            return jnp.where(mask, z, 2 * z ** 0.5 - 1), mask

        @jit
        def huber2(z, mask):
            rho1 = jnp.where(mask, 1, z ** -0.5)
            rho2 = jnp.where(mask, 0, -0.5 * z ** -1.5)
            return rho1, rho2

        self.huber1 = huber1
        self.huber2 = huber2

    def huber(self, z, cost_only):
        rho0, mask = self.huber1(z)
        if cost_only:
            rho1, rho2 = self.get_empty_rhos(z)
        else:
            rho1, rho2 = self.huber2(z, mask)
        return self.stack_rhos(rho0, rho1, rho2)

    def create_soft_l1_funcs(self):
        @jit
        def soft_l1_1(z):
            t = 1 + z
            return 2 * (t ** 0.5 - 1), t

        @jit
        def soft_l1_2(t):
            rho1 = t ** -0.5
            rho2 = -0.5 * t ** -1.5
            return rho1, rho2

        self.soft_l1_1 = soft_l1_1
        self.soft_l1_2 = soft_l1_2

    def soft_l1(self, z, cost_only):
        rho0, t = self.soft_l1_1(z)
        if cost_only:
            rho1, rho2 = self.get_empty_rhos(z)
        else:
            rho1, rho2 = self.soft_l1_2(t)
        return self.stack_rhos(rho0, rho1, rho2)

    def create_cauchy_funcs(self):
        @jit
        def cauchy1(z):
            return jnp.log1p(z)

        @jit
        def cauchy2(z):
            t = 1 + z
            rho1 = 1 / t
            rho2 = -1 / t ** 2
            return rho1, rho2

        self.cauchy1 = cauchy1
        self.cauchy2 = cauchy2

    def cauchy(self, z, cost_only):
        rho0 = self.cauchy1(z)
        if cost_only:
            rho1, rho2 = self.get_empty_rhos(z)
        else:
            rho1, rho2 = self.cauchy2(z)
        return self.stack_rhos(rho0, rho1, rho2)

    def create_arctan_funcs(self):
        @jit
        def arctan1(z):
            return jnp.arctan(z)

        @jit
        def arctan2(z):
            t = 1 + z ** 2
            return 1 / t, -2 * z / t ** 2

        self.arctan1 = arctan1
        self.arctan2 = arctan2

    def arctan(self, z, cost_only):
        rho0 = self.arctan1(z)
        if cost_only:
            rho1, rho2 = self.get_empty_rhos(z)
        else:
            rho1, rho2 = self.arctan2(z)
        return self.stack_rhos(rho0, rho1, rho2)

    def create_zscale(self):
        @jit
        def zscale(f, f_scale):
            return (f / f_scale) ** 2

        self.zscale = zscale

    def create_calculate_cost(self):
        @jit
        def calculate_cost(f_scale, rho, data_mask):
            cost_array = jnp.where(data_mask, rho[0], 0)
            return 0.5 * f_scale ** 2 * jnp.sum(cost_array)

        self.calculate_cost = calculate_cost

    def create_scale_rhos(self):
        @jit
        def scale_rhos(rho, f_scale):
            rho0 = rho[0] * f_scale ** 2
            rho1 = rho[1]
            rho2 = rho[2] / f_scale ** 2
            return self.stack_rhos(rho0, rho1, rho2)

        self.scale_rhos = scale_rhos

    def construct_single_loss_function(self, loss):
        def loss_function(f, f_scale, data_mask=None, cost_only=False):
            z = self.zscale(f, f_scale)
            rho = loss(z, cost_only=cost_only)
            if cost_only:
                return self.calculate_cost(f_scale, rho, data_mask)
            rho = self.scale_rhos(rho, f_scale)
            return rho

        return loss_function

    def construct_all_loss_functions(self):
        loss_funcs = {}
        for key, loss in self.IMPLEMENTED_LOSSES.items():
            loss_funcs[key] = self.construct_single_loss_function(loss)

        return loss_funcs

    def get_loss_function(self, loss):
        if loss == 'linear':
            return None

        if not callable(loss):
            return self.loss_funcs[loss]
        else:
            def loss_function(f, f_scale, data_mask=None, cost_only=False):
                z = self.zscale(f, f_scale)
                rho = loss(z)
                if cost_only:
                    return self.calculate_cost(f_scale, rho, data_mask)
                rho = self.scale_rhos(rho, f_scale)
                return rho

        return loss_function



"""Generic interface for least-squares minimization."""
from warnings import warn
import numpy as np
import time
from typing import Callable, Optional, Tuple, Union, Sequence, List, Any

from jax.config import config
config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax import jit, jacfwd
from jax.scipy.linalg import solve_triangular as jax_solve_triangular

from jaxfit.trf import TrustRegionReflective
from jaxfit.loss_functions import LossFunctionsJIT
from jaxfit.common_scipy import EPS, in_bounds, make_strictly_feasible


TERMINATION_MESSAGES = {
    -1: "Improper input parameters status returned from `leastsq`",
    0: "The maximum number of function evaluations is exceeded.",
    1: "`gtol` termination condition is satisfied.",
    2: "`ftol` termination condition is satisfied.",
    3: "`xtol` termination condition is satisfied.",
    4: "Both `ftol` and `xtol` termination conditions are satisfied."
}

def prepare_bounds(bounds, n) -> Tuple[np.ndarray, np.ndarray]:
    """Prepare bounds for optimization.

    This function prepares the bounds for the optimization by ensuring that 
    they are both 1-D arrays of length `n`. If either bound is a scalar, it is 
    resized to an array of length `n`.

    Parameters
    ----------
    bounds : Tuple[np.ndarray, np.ndarray]
        The lower and upper bounds for the optimization.
    n : int
        The length of the bounds arrays.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The prepared lower and upper bounds arrays.
    """
    lb, ub = [np.asarray(b, dtype=float) for b in bounds]
    if lb.ndim == 0:
        lb = np.resize(lb, n)

    if ub.ndim == 0:
        ub = np.resize(ub, n)

    return lb, ub


def check_tolerance(ftol: float, 
                    xtol: float, 
                    gtol: float, 
                    method: str
                    ) -> Tuple[float, float, float]:
    """Check and prepare tolerance values for optimization.

    This function checks the tolerance values for the optimization and 
    prepares them for use. If any of the tolerances is `None`, it is set to 
    0. If any of the tolerances is lower than the machine epsilon, a warning 
    is issued and the tolerance is set to the machine epsilon. If all 
    tolerances are lower than the machine epsilon, a `ValueError` is raised.

    Parameters
    ----------
    ftol : float
        The tolerance for the optimization function value.
    xtol : float
        The tolerance for the optimization variable values.
    gtol : float
        The tolerance for the optimization gradient values.
    method : str
        The name of the optimization method.

    Returns
    -------
    Tuple[float, float, float]
        The prepared tolerance values.
    """
    def check(tol: float, name: str) -> float:
        if tol is None:
            tol = 0
        elif tol < EPS:
            warn("Setting `{}` below the machine epsilon ({:.2e}) effectively "
                 "disables the corresponding termination condition."
                 .format(name, EPS))
        return tol

    ftol = check(ftol, "ftol")
    xtol = check(xtol, "xtol")
    gtol = check(gtol, "gtol")

    if ftol < EPS and xtol < EPS and gtol < EPS:
        raise ValueError("At least one of the tolerances must be higher than "
                         "machine epsilon ({:.2e}).".format(EPS))

    return ftol, xtol, gtol


def check_x_scale(x_scale: Union[str, Sequence[float]],
                  x0: Sequence[float]
                  ) -> Union[str, Sequence[float]]:
    """Check and prepare the `x_scale` parameter for optimization.

    This function checks and prepares the `x_scale` parameter for the 
    optimization. `x_scale` can either be 'jac' or an array_like with positive
    numbers. If it's 'jac' the jacobian is used as the scaling. 

    Parameters
    ----------
    x_scale : Union[str, Sequence[float]]
        The scaling for the optimization variables.
    x0 : Sequence[float]
        The initial guess for the optimization variables.

    Returns
    -------
    Union[str, Sequence[float]]
        The prepared `x_scale` parameter.
    """

    if isinstance(x_scale, str) and x_scale == 'jac':
        return x_scale

    try:
        x_scale = np.asarray(x_scale, dtype=float)
        valid = np.all(np.isfinite(x_scale)) and np.all(x_scale > 0)
    except (ValueError, TypeError):
        valid = False

    if not valid:
        raise ValueError("`x_scale` must be 'jac' or array_like with "
                         "positive numbers.")

    if x_scale.ndim == 0:
        x_scale = np.resize(x_scale, x0.shape)

    if x_scale.shape != x0.shape:
        raise ValueError("Inconsistent shapes between `x_scale` and `x0`.")

    return x_scale

"""Wraps the given function such that a masked jacfwd is performed on it
thereby giving the autodiff jacobian."""
class AutoDiffJacobian():
    """Wraps the residual fit function such that a masked jacfwd is performed 
    on it. thereby giving the autodiff Jacobian. This needs to be a class since 
    we need to maintain in memory three different versions of the Jacobian.
    """

    def create_ad_jacobian(self, 
                           func: Callable, 
                           num_args: int, 
                           masked: bool = True
                           ) -> Callable:
        """Creates a function that returns the autodiff jacobian of the 
        residual fit function. The Jacobian of the residual fit function is
        equivalent to the Jacobian of the fit function.

        Parameters
        ----------
        func : Callable
            The function to take the jacobian of.
        num_args : int
            The number of arguments the function takes.
        masked : bool, optional
            Whether to use a masked jacobian, by default True

        Returns
        -------
        Callable
            The function that returns the autodiff jacobian of the given
            function.
        """

        # create a list of argument indices for the wrapped function which
        # will correspond to the arguments of the residual fit function and
        # will be past to JAX's jacfwd function.
        arg_list = [4 + i for i in range(num_args)]
            
        @jit
        def wrap_func(*all_args: List[Any]) -> jnp.ndarray:
            """Wraps the residual fit function such that it can be passed to the
            jacfwd function. Jacfwd requires the function to a single list
            of arguments.
            """
            xdata, ydata, data_mask, atransform = all_args[:4]
            args = all_args[4:]
            return func(args, xdata, ydata, data_mask, atransform)
        
        @jit
        def jac_func(args: List[float], 
                     xdata: jnp.ndarray, 
                     ydata: jnp.ndarray, 
                     data_mask: jnp.ndarray, 
                     atransform: jnp.ndarray
                     ) -> jnp.ndarray:
            """Returns the jacobian. Places all the residual fit function
            arguments into a single list for the wrapped residual fit function.
            Then calls the jacfwd function on the wrapped function with the
            the arglist of the arguments to differentiate with respect to which
            is only the arguments of the original fit function.
            """

            fixed_args = [xdata, ydata, data_mask, atransform] 
            all_args = [*fixed_args, *args]
            jac_fwd = jacfwd(wrap_func, argnums=arg_list)(*all_args)
            return jnp.array(jac_fwd)
        
        @jit
        def masked_jac(args: List[float], 
                       xdata: jnp.ndarray, 
                       ydata: jnp.ndarray, 
                       data_mask: jnp.ndarray, 
                       atransform: jnp.ndarray
                       ) -> jnp.ndarray:
            """Returns the masked jacobian."""
            Jt = jac_func(args, xdata, ydata, data_mask, atransform)
            J = jnp.where(data_mask, Jt, 0).T 
            return jnp.atleast_2d(J)
         
        @jit
        def no_mask_jac(args: List[float],
                        xdata: jnp.ndarray, 
                        ydata: jnp.ndarray,
                        data_mask: jnp.ndarray, 
                        atransform: jnp.ndarray
                        ) -> jnp.ndarray:
            """Returns the unmasked jacobian."""
            J = jac_func(args, xdata, ydata, data_mask, atransform).T
            return jnp.atleast_2d(J)
        
        if masked:
            self.jac = masked_jac
        else:
            self.jac = no_mask_jac
        return self.jac
    


class LeastSquares():
    
    def __init__(self):
        super().__init__() # not sure if this is needed
        self.trf = TrustRegionReflective()
        self.ls = LossFunctionsJIT()
        #initialize jacobian to None and f to a dummy function
        self.f = lambda x: None 
        self.jac = None
        
        # need a separate instance of the autodiff class for each of the 
        # the different sigma/covariance cases
        self.adjn = AutoDiffJacobian()
        self.adj1d = AutoDiffJacobian()
        self.adj2d = AutoDiffJacobian()
    

    def least_squares(self, 
                      fun: Callable, 
                      x0: np.ndarray, 
                      jac: Optional[Callable] = None, 
                      bounds: Tuple[np.ndarray, np.ndarray] = (-np.inf, np.inf),
                      method: str = 'trf',
                      ftol: float = 1e-8,
                      xtol: float = 1e-8, 
                      gtol: float = 1e-8, 
                      x_scale: Union[str, np.ndarray, float] = 1.0,
                      loss: str = 'linear',
                      f_scale: float = 1.0,
                      diff_step=None, 
                      tr_solver=None, 
                      tr_options={},
                      jac_sparsity=None, 
                      max_nfev: Optional[float] = None, 
                      verbose: int = 0, 
                      xdata: Optional[jnp.ndarray] = None, 
                      ydata: Optional[jnp.ndarray] = None, 
                      data_mask: Optional[jnp.ndarray] = None, 
                      transform: Optional[jnp.ndarray] = None, 
                      timeit: bool = False,
                      args=(), 
                      kwargs={}):
        
        
        if data_mask is None and ydata is not None:
            data_mask = jnp.ones(len(ydata), dtype=bool)
            
        if loss not in self.ls.IMPLEMENTED_LOSSES and not callable(loss):
            raise ValueError("`loss` must be one of {0} or a callable."
                             .format(self.ls.IMPLEMENTED_LOSSES.keys()))
        
        if method not in ['trf']:
            raise ValueError("`method` must be 'trf")
            
        if jac not in [None] and not callable(jac):
            raise ValueError("`jac` must be None or "
                             "callable.")
    
        if verbose not in [0, 1, 2]:
            raise ValueError("`verbose` must be in [0, 1, 2].")
    
        if len(bounds) != 2:
            raise ValueError("`bounds` must contain 2 elements.")
    
        if max_nfev is not None and max_nfev <= 0:
            raise ValueError("`max_nfev` must be None or positive integer.")
    
        if np.iscomplexobj(x0):
            raise ValueError("`x0` must be real.")
    
        x0 = np.atleast_1d(x0).astype(float)
    
        if x0.ndim > 1:
            raise ValueError("`x0` must have at most 1 dimension.")
            
        self.n = len(x0)

    
        lb, ub = prepare_bounds(bounds, x0.shape[0])
    
        if lb.shape != x0.shape or ub.shape != x0.shape:
            raise ValueError("Inconsistent shapes between bounds and `x0`.")
    
        if np.any(lb >= ub):
            raise ValueError("Each lower bound must be strictly less than each "
                             "upper bound.")
    
        if not in_bounds(x0, lb, ub):
            raise ValueError("`x0` is infeasible.")
            
    
        x_scale = check_x_scale(x_scale, x0)
        ftol, xtol, gtol = check_tolerance(ftol, xtol, gtol, method)
        x0 = make_strictly_feasible(x0, lb, ub)
        

        if xdata is not None and ydata is not None:
            # checks to see if the fit function is the same. Can't directly
            # compare the functions so we compare function code directly
            func_update = self.f.__code__.co_code != fun.__code__.co_code
            # if we are updating the fit function then we need to update the
            # jacobian function as well
            if func_update:
                self.update_function(fun)
                # this only updates the the jacobian if using autodiff (jac=None)
                if jac is None:
                    self.autdiff_jac(jac)
                    
            # if using an analytical jacobian
            if jac is not None:
                # if we are in the first function call
                if self.jac is None:
                    self.wrap_jac(jac)
                elif self.jac.__code__.co_code != jac.__code__.co_code:
                    # checks to see if the jacobian function is the same (see
                    # func_update for why no direct comparing of the functions)
                    # if it's a different Jacobian we need to rewrap it
                    self.wrap_jac(jac)
            elif self.jac is not None and not func_update:
                self.autdiff_jac(jac)
        
            # determines the correct residual function and jacobian to use
            # depending on whether data uncertainty transform is None, 1D, or 2D
            if transform is None:
                rfunc = self.func_none
                jac_func = self.jac_none
            elif transform.ndim == 1:
                rfunc = self.func_1d
                jac_func = self.jac_1d
            else:
                rfunc = self.func_2d
                jac_func = self.jac_2d
        else:
            # this if/else is to maintain compatibility with the SciPy suite of tests
            # which assume the residual function contains the fit data which is not
            # the case for JAXFit due to how we've made the residual function
            # function to be compatible with JAX JIT compilation
            def wrap_func(fargs, xdata, ydata, data_mask, atransform):
                return jnp.atleast_1d(fun(fargs, *args, **kwargs))
            
            def wrap_jac(fargs, xdata, ydata, data_mask, atransform):
                return jnp.atleast_2d(jac(fargs, *args, **kwargs))
            
            rfunc = wrap_func
            if jac is None:
                adj = AutoDiffJacobian()
                jac_func = adj.create_ad_jacobian(wrap_func, self.n, masked=False)
            else:
                jac_func = wrap_jac
            
                
                
        f0 = rfunc(x0, xdata, ydata, data_mask, transform)
        J0 = jac_func(x0, xdata, ydata, data_mask, transform)

        if f0.ndim != 1:
            raise ValueError("`fun` must return at most 1-d array_like. "
                             "f0.shape: {0}".format(f0.shape))

        if not np.all(np.isfinite(f0)):
            raise ValueError("Residuals are not finite in the initial point.")

        n = x0.size
        m = f0.size
        
        if J0 is not None:
            if J0.shape != (m, n):
                raise ValueError(
                    "The return value of `jac` has wrong shape: expected {0}, "
                    "actual {1}.".format((m, n), J0.shape))
                
        if data_mask is None:
            data_mask = jnp.ones(m)
        
        loss_function = self.ls.get_loss_function(loss)

        if callable(loss):
            rho = loss_function(f0, f_scale, data_mask=data_mask)
            if rho.shape != (3, m):
                raise ValueError("The return value of `loss` callable has wrong "
                                 "shape.")
            initial_cost_jnp = self.trf.calculate_cost(rho, data_mask)
        elif loss_function is not None:
            initial_cost_jnp = loss_function(f0, f_scale, data_mask=data_mask, 
                                             cost_only=True)
        else:
            initial_cost_jnp = self.trf.default_loss_func(f0)
        initial_cost = np.array(initial_cost_jnp)
        
        result = self.trf.trf(rfunc, xdata, ydata, jac_func, data_mask, 
                              transform, x0, f0, J0, lb, ub, ftol, xtol,
                     gtol, max_nfev, f_scale, x_scale, loss_function,
                     tr_options.copy(), verbose, timeit)
    
    
        result.message = TERMINATION_MESSAGES[result.status]
        result.success = result.status > 0
    
        if verbose >= 1:
            print(result.message)
            print("Function evaluations {0}, initial cost {1:.4e}, final cost "
                  "{2:.4e}, first-order optimality {3:.2e}."
                  .format(result.nfev, initial_cost, result.cost,
                          result.optimality))
            
        return result
   
            
    def autdiff_jac(self, jac: None) -> None:
        """We do this for all three sigma transformed functions such
        that if sigma is changed from none to 1D to covariance sigma then no
        retracing is needed.

        Parameters
        ----------
        jac : None
            Passed in to maintain compatibility with the user defined Jacobian
            function.
        """
        self.jac_none = self.adjn.create_ad_jacobian(self.func_none, self.n)
        self.jac_1d = self.adj1d.create_ad_jacobian(self.func_1d, self.n)
        self.jac_2d = self.adj2d.create_ad_jacobian(self.func_2d, self.n)
        # jac is
        self.jac = jac


    def update_function(self, func: Callable) -> None:
        """Wraps the given fit function to be a residual function using the
        data. The wrapped function is in a JAX JIT compatible format which
        is purely functional. This requires that both the data mask and the
        uncertainty transform are passed to the function. Even for the case
        where the data mask is all True and the uncertainty transform is None
        we still need to pass these arguments to the function due JAX's
        functional nature.

        Parameters
        ----------
        func : Callable
            The fit function to wrap.

        Returns
        -------
        None
        """
 
        @jit
        def masked_residual_func(args: List[float], 
                                 xdata: jnp.ndarray, 
                                 ydata: jnp.ndarray, 
                                 data_mask: jnp.ndarray
                                 ) -> jnp.ndarray:

            """Compute the residual of the function evaluated at `args` with 
            respect to the data.

            This function computes the residual of the user fit function 
            evaluated at `args` with respect to the data `(xdata, ydata)`, 
            masked by `data_mask`. The residual is defined as the difference 
            between the function evaluation and the data. The masked residual 
            is obtained by setting the residual to 0 wherever the corresponding 
            element of `data_mask` is 0.

            Parameters
            ----------
            args : List[float]
                The parameters of the function.
            xdata : jnp.ndarray
                The independent variable data.
            ydata : jnp.ndarray
                The dependent variable data.
            data_mask : jnp.ndarray
                The mask for the data.

            Returns
            -------
            jnp.ndarray
                The masked residual of the function evaluated at `args` with respect to the data.
            """
            func_eval = func(xdata, *args) - ydata
            return jnp.where(data_mask, func_eval, 0)

        # need to define a separate function for each of the different
        # sigma/covariance cases as the uncertainty transform is different
        # for each case. In future could remove the no transfore bit by setting
        # the uncertainty transform to all ones in the case where there is no
        # uncertainty transform.
        
        @jit
        def func_no_transform(args: List[float], 
                              xdata: jnp.ndarray, 
                              ydata: jnp.ndarray, 
                              data_mask: jnp.ndarray, 
                              atransform: jnp.ndarray
                              ) -> jnp.ndarray:

            """The residual function when there is no uncertainty transform.
            The atranform argument is not used in this case, but is included
            for consistency with the other cases."""
            return masked_residual_func(args, xdata, ydata, data_mask)
        
        @jit
        def func_1d_transform(args: List[float], 
                              xdata: jnp.ndarray, 
                              ydata: jnp.ndarray, 
                              data_mask: jnp.ndarray, 
                              atransform: jnp.ndarray
                              ) -> jnp.ndarray:
            """The residual function when there is a 1D uncertainty transform,
            that is when only the diagonal elements of the inverse covariance
            matrix are used."""
            return atransform * masked_residual_func(args, xdata, 
                                                     ydata, data_mask)
        
        @jit
        def func_2d_transform(args: List[float], 
                              xdata: jnp.ndarray, 
                              ydata: jnp.ndarray, 
                              data_mask: jnp.ndarray, 
                              atransform: jnp.ndarray
                              ) -> jnp.ndarray:
            """The residual function when there is a 2D uncertainty transform,
            that is when the full covariance matrix is given."""
            f = masked_residual_func(args, xdata, ydata, data_mask)
            return jax_solve_triangular(atransform, f, lower=True)
        
        self.func_none = func_no_transform
        self.func_1d = func_1d_transform
        self.func_2d = func_2d_transform
        self.f = func
    

    def wrap_jac(self, jac: Callable) -> None:
        """Wraps an user defined Jacobian function to allow for data masking
        and uncertainty transforms. The wrapped function is in a JAX JIT
        compatible format which is purely functional. This requires that both
        the data mask and the uncertainty transform are passed to the function.
        
        Using an analytical Jacobian of the fit function is equivalent to
        the Jacobian of the residual function.

        Also note that the analytical Jacobian doesn't require the independent
        ydata, but we still need to pass it to the function to maintain
        compatibility with autdiff version which does require the ydata.
        
        Parameters
        ----------
        jac : Callable
            The Jacobian function to wrap.
            
        Returns
        -------
        jnp.ndarray
            The masked Jacobian of the function evaluated at `args` with respect to the data.
        """
        
        @jit
        def jac_func(coords: jnp.ndarray,
                     args: List[float]
                     ) -> jnp.ndarray:
            jac_fwd = jac(coords, *args)
            return jnp.array(jac_fwd)
        
        @jit
        def masked_jac(coords: jnp.ndarray, 
                       args: List[float], 
                       data_mask: jnp.ndarray
                       ) -> jnp.ndarray:
             """Compute the wrapped Jacobian but masks out the padded elements 
             with 0s"""
             Jt = jac_func(coords, args)
             return jnp.where(data_mask, Jt, 0).T        
        
        @jit
        def jac_no_transform(args: List[float], 
                             coords: jnp.ndarray, 
                             ydata: jnp.ndarray, 
                             data_mask: jnp.ndarray, 
                             atransform: jnp.ndarray
                             ) -> jnp.ndarray:
            """The wrapped Jacobian function when there is no 
            uncertainty transform."""
            return jnp.atleast_2d(masked_jac(coords, args, data_mask))

        @jit
        def jac_1d_transform(args: List[float], 
                             coords: jnp.ndarray, 
                             ydata: jnp.ndarray, 
                             data_mask: jnp.ndarray, 
                             atransform: jnp.ndarray
                             ) -> jnp.ndarray:
            """The wrapped Jacobian function when there is a 1D uncertainty
            transform, that is when only the diagonal elements of the inverse
            covariance matrix are used."""
            J = masked_jac(coords, args, data_mask)
            return jnp.atleast_2d(atransform[:, jnp.newaxis] * jnp.asarray(J))
        
        @jit
        def jac_2d_transform(args: List[float], 
                             coords: jnp.ndarray, 
                             ydata: jnp.ndarray, 
                             data_mask: jnp.ndarray, 
                             atransform: jnp.ndarray
                             ) -> jnp.ndarray:
            """The wrapped Jacobian function when there is a 2D uncertainty
            transform, that is when the full covariance matrix is given."""

            J = masked_jac(coords, args, data_mask)
            return jnp.atleast_2d(jax_solve_triangular(atransform, 
                                                       jnp.asarray(J), 
                                                       lower=True))
        # we need all three versions of the Jacobian function to allow for
        # changing the sigma transform from none to 1D to 2D without having
        # to retrace the function
        self.jac_none = jac_no_transform
        self.jac_1d = jac_1d_transform
        self.jac_2d = jac_2d_transform
        self.jac = jac


# -*- coding: utf-8 -*-
"""
Created on Mon Mar 14 19:26:59 2022

@author: hofer
"""


from .minpack import *
from .least_squares import *
from .loss_functions import *
from .trf import *
from .common_jax import *
from .common_scipy import *
from ._optimize import *

__all__ = [s for s in dir() if not s.startswith('_')]


import warnings


class OptimizeResult(dict):
    """ Represents the optimization result.
    Attributes
    ----------
    x : ndarray
        The solution of the optimization.
    success : bool
        Whether or not the optimizer exited successfully.
    status : int
        Termination status of the optimizer. Its value depends on the
        underlying solver. Refer to `message` for details.
    message : str
        Description of the cause of the termination.
    fun, jac, hess: ndarray
        Values of objective function, its Jacobian and its Hessian (if
        available). The Hessians may be approximations, see the documentation
        of the function in question.
    hess_inv : object
        Inverse of the objective function's Hessian; may be an approximation.
        Not available for all solvers. The type of this attribute may be
        either np.ndarray or scipy.sparse.linalg.LinearOperator.
    nfev, njev, nhev : int
        Number of evaluations of the objective functions and of its
        Jacobian and Hessian.
    nit : int
        Number of iterations performed by the optimizer.
    maxcv : float
        The maximum constraint violation.
    Notes
    -----
    `OptimizeResult` may have additional attributes not listed here depending
    on the specific solver being used. Since this class is essentially a
    subclass of dict with attribute accessors, one can see which
    attributes are available using the `OptimizeResult.keys` method.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(name) from e

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])
        else:
            return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())
    
    
def _check_unknown_options(unknown_options):
    if unknown_options:
        msg = ", ".join(map(str, unknown_options.keys()))
        # Stack level 4: this is called from _minimize_*, which is
        # called from another function in SciPy. Level 4 is the first
        # level in user code.
        warnings.warn("Unknown solver options: %s" % msg, OptimizeWarning, 4)
        
        
class OptimizeWarning(UserWarning):
    pass